/* * This file is part of Harmattan Organiser GUI *
 *
 * Copyright (C) 2009-2010 Nokia Corporation and/or its subsidiary(-ies).
 * All rights reserved.
 *
 * Contact: Andrey Moiseenko <andrey.moiseenko@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */


#include "defaultinvitationplugin.h"
#include <extendedcalendar.h>
#include <QDebug>
#include <icalformat.h>
#include <QTimer>

#ifdef MKCAL_FOR_MEEGO
#include <QMailAccount>
#include <QMailStore>
#include <QMailAddress>
#include <QMailBase64Codec>
#include <qmailnamespace.h>

#include <QFile>

#include "transmitemail.h"

#endif

using namespace KCalCore;

const QString name("DefaultInvitationPlugin");

//@cond PRIVATE
class DefaultInvitationPlugin::Private
{
public:
  Private() : mStore( 0 ), mInit ( false ),
  mErrorCode( ServiceInterface::ErrorOk )
  {
  }
  ~Private() {
    uninit();
  }

  bool init() {

    if ( !mInit ) {

      const QString& path = QMail::messageServerLockFilePath();
      if (!QFile::exists(path)) {
        mErrorCode = ServiceInterface::ErrorNoAccount;
        qWarning() << "Message server not running, aborting.";
        return false;
      }

      mStore = QMailStore::instance();
      Q_ASSERT(mStore);

      mInit = true;
    }
    return mInit;
  }

  bool isInit() {
      return mInit;
  }

  QMailAccount* getAccount() {

      if ( mInit ) {
          QMailAccountKey byDefault = QMailAccountKey::status( QMailAccount::PreferredSender );
          QMailAccountIdList accounts = mStore->queryAccounts(byDefault);
          if (!accounts.count()) {
              qWarning() << "Default account was not found!";
              mErrorCode = ServiceInterface::ErrorNoAccount;
              return 0;
          }
          if (accounts.count() > 1) {
              qWarning( "There are more than one default account, using first" );
          }
          return new QMailAccount( accounts.first() );

      } else {
          mErrorCode = ServiceInterface::ErrorNoAccount;
          return 0;
      }
  }

  QString defaultAddress() {

    QString result;

    if ( mInit ) {
      const QMailAccount* account = getAccount();
      if (account) {
          result = account->fromAddress().address();
          delete account;
      }
    }
    return result;
  }


  void uninit() {

    delete mStore;
    mStore = 0;
    mInit = false;
  }

  bool sendMail( const QStringList &recipients, const QString &subject,
                 const QString &body, const QString &attachment ) {

    if ( !mInit ) {
      return false;
    }


    // Build a message
    QMailMessage message;

    const QMailAccount* account = getAccount();
    if ( account ) {

        // Setup account which should be used to send a message
        message.setParentAccountId( account->id() );
        // Define from address
        message.setFrom( account->fromAddress() );

        // Get the outbox folder ID
        QMailFolderId folderId = account->standardFolder(QMailFolder::OutboxFolder);
        if (!folderId.isValid()) {
            folderId = QMailFolder::LocalStorageFolderId;
        }
        // Put message to standard outbox folder fo that account
        message.setParentFolderId( folderId );


        // Setup message status
        message.setStatus(QMailMessage::Outbox, true);
        message.setStatus(QMailMessage::Outgoing, true);
        message.setStatus(QMailMessage::ContentAvailable, true);
        message.setStatus(QMailMessage::PartialContentAvailable, true);
        message.setStatus(QMailMessage::Read, true);
        message.setDate(QMailTimeStamp(QDateTime::currentDateTime()));

        // Define recipeint's address
        QList<QMailAddress> addresses;
        foreach( const QString &mail, recipients) {
          addresses.append( QMailAddress( mail ) );
        }
        message.setTo( addresses );
        // Define subject
        message.setSubject( subject );
        message.setMessageType(QMailMessage::Email);
        message.setMultipartType(QMailMessagePartContainerFwd::MultipartAlternative);

        // Create the MIME part representing the message body
        QMailMessagePart bodyPart = QMailMessagePart::fromData(
            body,
            QMailMessageContentDisposition(QMailMessageContentDisposition::None),
            QMailMessageContentType("text/plain; charset=\"utf-8\""),
            QMailMessageBody::QuotedPrintable);
        bodyPart.removeHeaderField("Content-Disposition");

        // Create the calendar MIME part
        QMailMessagePart calendarPart = QMailMessagePart::fromData(
            attachment,
            QMailMessageContentDisposition(QMailMessageContentDisposition::None),
            QMailMessageContentType("text/calendar; charset=\"utf-8\"; method=REQUEST"),
            QMailMessageBody::Base64);
        calendarPart.removeHeaderField("Content-Disposition");
        calendarPart.appendHeaderField("Content-Class","urn:content-classes:calendarmessage");

        message.appendPart(bodyPart);
        message.appendPart(calendarPart);

        // send (to outbox)
        if (!mStore->addMessage(&message)) {
          delete account;
          return false;
        }

        // initiate transmission
        QSharedPointer<QMailTransmitAction> action ( new QMailTransmitAction() );
        if ( action ) {
          action->transmitMessages( account->id() );
        } else {
          qDebug( "No MailTransmistAction" );
          delete account;
          return false;
        }

        delete account;
    } else {
        mErrorCode = ServiceInterface::ErrorNoAccount;
        return false;
    }


    return true;
  }

  QMailStore *mStore;

  bool mInit;
  ServiceInterface::ErrorCode mErrorCode;
};


DefaultInvitationPlugin::DefaultInvitationPlugin(): d( new Private() )
{
}

DefaultInvitationPlugin::~DefaultInvitationPlugin()
{
  delete d;
}


bool DefaultInvitationPlugin::sendInvitation(const QString &accountId, const QString &notebookUid, const Incidence::Ptr &invitation, const QString &body)
{

  Q_UNUSED( body );
  Q_UNUSED( accountId );
  Q_UNUSED( notebookUid );

  if ( !d->init() ) {
      d->mErrorCode = ServiceInterface::ErrorNoAccount;
      return false;
  }

  d->mErrorCode = ServiceInterface::ErrorOk;

  QStringList emails;
  const Incidence::Ptr &clonedInvitation = Incidence::Ptr(invitation->clone());
  const Attendee::List &attendees = clonedInvitation->attendees();
  foreach (const Attendee::Ptr &attendee, attendees) {
      if (Attendee::Removed != attendee->status()) {
          emails.append(attendee->email());
      }
      else {
          clonedInvitation->deleteAttendee(attendee);
      }
  }

  // check if we have any recipients
  if (emails.empty()) {
      qDebug("No attendees");
      return false;
  }

  // Reset the alternative description
  clonedInvitation->setAltDescription(QString::null);

  ICalFormat icf;
  const QString &ical =  icf.createScheduleMessage(clonedInvitation, iTIPRequest);
  const QString &description = clonedInvitation->description();
  const bool res = d->sendMail(emails, clonedInvitation->summary(), description, ical);

//  d->uninit();
  return res;
}

bool DefaultInvitationPlugin::sendUpdate(const QString &accountId, const Incidence::Ptr &invitation, const QString &body)
{
  //Same as sendResponse
  return sendResponse(accountId, invitation, body);
}

bool DefaultInvitationPlugin::sendResponse(const QString &accountId, const Incidence::Ptr &invitation, const QString &body)
{
  Q_UNUSED(accountId);    // not needed by this plugin: we use the default account

  if ( !d->init() ) {
      d->mErrorCode = ServiceInterface::ErrorNoAccount;
      return false;
  }

  d->mErrorCode = ServiceInterface::ErrorOk;

  d->init();
  // Is there an organizer?
  Person::Ptr organizer = invitation->organizer();
  if (organizer->isEmpty() || organizer->email().isEmpty()) { // we do not have an organizer
    qWarning() << "sendResponse() called with wrong invitation: there is no organizer!";
    return false;
  }

  // Check: Am I one of the attendees? Had the organizer requested RSVP from me?
  Attendee::Ptr me = invitation->attendeeByMail( d->defaultAddress() );
  if (me == 0 || !me->RSVP()) {
    qWarning() << "sendResponse() called with wrong invitation: we are not invited or no response is expected.";
    return false;
  }

  ICalFormat icf;
  QString ical =  icf.createScheduleMessage( invitation, iTIPReply) ;
  //    QString base64Data = ical;

  bool res = d->sendMail(QStringList( organizer->email() ), invitation->summary(), body, ical);

//  d->uninit();
  return res;

}

QString DefaultInvitationPlugin::pluginName() const
{
  d->mErrorCode = ServiceInterface::ErrorOk;
  return name;
}

QString DefaultInvitationPlugin::icon() const
{
  return QString();
}

QString DefaultInvitationPlugin::uiName() const
{
  return QLatin1String("Default");
}

bool DefaultInvitationPlugin::multiCalendar() const
{
  d->mErrorCode = ServiceInterface::ErrorNotSupported;
  return false;
}

QString DefaultInvitationPlugin::emailAddress(const mKCal::Notebook::Ptr &notebook)
{
  Q_UNUSED( notebook );
  QString email;

  qDebug() << "Requesting the email from plugin!!!";

  if ( d->init() ) {
    email = d->defaultAddress();
    d->mErrorCode = ServiceInterface::ErrorOk;
  } else {
    d->mErrorCode = ServiceInterface::ErrorNoAccount;
  }

  return email;
}

QString DefaultInvitationPlugin::displayName(const mKCal::Notebook::Ptr &notebook) const
{
  Q_UNUSED( notebook );
  return QString();
}

bool DefaultInvitationPlugin::downloadAttachment(const mKCal::Notebook::Ptr &notebook, const QString &uri, const QString &path)
{
  Q_UNUSED( notebook );
  Q_UNUSED( uri );
  Q_UNUSED( path );
  d->mErrorCode = ServiceInterface::ErrorNotSupported;
  return false;

}

bool DefaultInvitationPlugin::deleteAttachment(const mKCal::Notebook::Ptr &notebook, const Incidence::Ptr &incidence, const QString &uri)
{
  Q_UNUSED( notebook );
  Q_UNUSED( incidence );
  Q_UNUSED( uri );
  d->mErrorCode = ServiceInterface::ErrorNotSupported;
  return false;

}

bool DefaultInvitationPlugin::shareNotebook(const mKCal::Notebook::Ptr &notebook, const QStringList &sharedWith)
{
  Q_UNUSED( notebook );
  Q_UNUSED( sharedWith );
  d->mErrorCode = ServiceInterface::ErrorNotSupported;
  return false;
}

QStringList DefaultInvitationPlugin::sharedWith(const mKCal::Notebook::Ptr &notebook)
{
  Q_UNUSED( notebook );
  d->mErrorCode = ServiceInterface::ErrorNotSupported;
  return QStringList();
}

QString DefaultInvitationPlugin::DefaultInvitationPlugin::serviceName() const
{
  d->mErrorCode = ServiceInterface::ErrorOk;
  return name;
}


QString DefaultInvitationPlugin::defaultNotebook() const
{
  d->mErrorCode = ServiceInterface::ErrorNotSupported;
  return QString();
}

bool DefaultInvitationPlugin::checkProductId(const QString &prodId) const
{
  Q_UNUSED(prodId);
  d->mErrorCode = ServiceInterface::ErrorNotSupported;
  return false;
}

ServiceInterface::ErrorCode DefaultInvitationPlugin::error() const
{
  return d->mErrorCode;
}


Q_EXPORT_PLUGIN2(defaultinvitationplugin, DefaultInvitationPlugin);
