/*
  This file is part of the mkcal library.

  Copyright (c) 2002,2003 Cornelius Schumacher <schumacher@kde.org>
  Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies). All rights reserved.
  Contact: Alvaro Manera <alvaro.manera@nokia.com>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; see the file COPYING.LIB.  If not, write to
  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
  Boston, MA 02110-1301, USA.
*/
/**
  @file
  This file is part of the API for handling calendar data and
  defines the ExtendedStorage abstract base class.

  @brief
  An abstract base class that provides a calendar storage interface.

  @author Cornelius Schumacher \<schumacher@kde.org\>
*/
#include <config-mkcal.h>
#include "extendedstorage.h"

#include <exceptions.h>
#include <calendar.h>
using namespace KCalCore;

#include <kdebug.h>

#if defined(HAVE_UUID_UUID_H)
#include <uuid/uuid.h>
#endif

#if defined(MKCAL_FOR_MEEGO)
#include <QtCore/QMap>
#include <QtDBus/QDBusConnection>
#include <QtDBus/QDBusInterface>
#include <QtDBus/QDBusReply>
#endif

#ifdef TIMED_SUPPORT
#include <timed/interface>
#include <timed/event>
#include <timed/exception>
using namespace Maemo;
#endif

using namespace mKCal;

/**
  Private class that helps to provide binary compatibility between releases.
  @internal
*/
//@cond PRIVATE
class mKCal::ExtendedStorage::Private
{
  public:
    Private( const ExtendedCalendar::Ptr &cal, bool validateNotebooks )
      : mCalendar( cal ),
        mValidateNotebooks( validateNotebooks ),
        mIsUncompletedTodosLoaded( false ),
        mIsCompletedTodosDateLoaded( false ),
        mIsCompletedTodosCreatedLoaded( false ),
        mIsDateLoaded( false ),
        mIsCreatedLoaded( false ),
        mIsFutureDateLoaded( false ),
        mIsGeoDateLoaded( false ),
        mIsGeoCreatedLoaded( false ),
        mIsUnreadIncidencesLoaded ( false ),
        mIsInvitationIncidencesLoaded ( false ),
        mIsJournalsLoaded( false ),
        mDefaultNotebook( 0 ),
	mIsOwnSave( false )
    {}

    ExtendedCalendar::Ptr mCalendar;
    bool mValidateNotebooks;
    QDate mStart;
    QDate mEnd;
    bool mIsUncompletedTodosLoaded;
    bool mIsCompletedTodosDateLoaded;
    bool mIsCompletedTodosCreatedLoaded;
    bool mIsDateLoaded;
    bool mIsCreatedLoaded;
    bool mIsFutureDateLoaded;
    bool mIsGeoDateLoaded;
    bool mIsGeoCreatedLoaded;
    bool mIsUnreadIncidencesLoaded;
    bool mIsInvitationIncidencesLoaded;
    bool mIsJournalsLoaded;
    QList<ExtendedStorageObserver*> mObservers;
    QHash<QString,Notebook::Ptr> mNotebooks; // uid to notebook
    Notebook::Ptr mDefaultNotebook;
    bool mIsOwnSave;

#if defined(TIMED_SUPPORT)
    int setAlarms( const Incidence::Ptr &incidence, Timed::Event::List &events, const KDateTime &now);
#endif
};
//@endcond

ExtendedStorage::ExtendedStorage( const ExtendedCalendar::Ptr &cal, bool validateNotebooks )
  : CalStorage( cal ),
    d( new ExtendedStorage::Private ( cal, validateNotebooks ) )
{
  // Add the calendar as observer
  registerObserver( cal.data() );
}

ExtendedStorage::~ExtendedStorage()
{
  delete d;
}

void ExtendedStorage::clearLoaded(bool mySave)
{
  if (mySave)
    return;

  d->mStart = QDate();
  d->mEnd = QDate();
  d->mIsUncompletedTodosLoaded = false;
  d->mIsCompletedTodosDateLoaded = false;
  d->mIsCompletedTodosCreatedLoaded = false;
  d->mIsDateLoaded = false;
  d->mIsCreatedLoaded = false;
  d->mIsFutureDateLoaded = false;
  d->mIsGeoDateLoaded = false;
  d->mIsGeoCreatedLoaded = false;
  d->mIsUnreadIncidencesLoaded  = false;
  d->mIsInvitationIncidencesLoaded  = false;
  d->mIsJournalsLoaded = false;
}

bool ExtendedStorage::getLoadDates( const QDate &start, const QDate &end,
                                    KDateTime &loadStart, KDateTime &loadEnd )
{
  // Check the need to load from db.
  if ( start.isValid() && d->mStart.isValid() && start >= d->mStart &&
       end.isValid() && d->mEnd.isValid() && end <= d->mEnd ) {
    return false;
  }

  // Set load dates to load only what's necessary.
  if ( start.isValid() && d->mStart.isValid() && start >= d->mStart ) {
    loadStart.setDate( d->mEnd );
  } else {
    loadStart.setDate( start ); // may be null if start is not valid
  }

  if ( end.isValid() && d->mEnd.isValid() && end <= d->mEnd ) {
    loadEnd.setDate( d->mStart );
  } else {
    loadEnd.setDate( end ); // may be null if end is not valid
  }

  loadStart.setTimeSpec( calendar()->timeSpec() );
  loadEnd.setTimeSpec( calendar()->timeSpec() );

  kDebug() << "get load dates" << start << end << loadStart << loadEnd;

  return true;
}

void ExtendedStorage::setLoadDates( const QDate &start, const QDate &end )
{
  // Set dates.
  if ( start.isValid() && ( !d->mStart.isValid() || start < d->mStart ) ) {
    d->mStart = start;
  }
  if ( end.isValid() && ( !d->mEnd.isValid() || end > d->mEnd ) ) {
    d->mEnd = end;
  }

  kDebug() << "set load dates" << d->mStart << d->mEnd;
}

bool ExtendedStorage::isUncompletedTodosLoaded()
{
  return d->mIsUncompletedTodosLoaded;
}

void ExtendedStorage::setIsUncompletedTodosLoaded( bool loaded )
{
  d->mIsUncompletedTodosLoaded = loaded;
}

bool ExtendedStorage::isCompletedTodosDateLoaded()
{
  return d->mIsCompletedTodosDateLoaded;
}

void ExtendedStorage::setIsCompletedTodosDateLoaded( bool loaded )
{
  d->mIsCompletedTodosDateLoaded = loaded;
}

bool ExtendedStorage::isCompletedTodosCreatedLoaded()
{
  return d->mIsCompletedTodosCreatedLoaded;
}

void ExtendedStorage::setIsCompletedTodosCreatedLoaded( bool loaded )
{
  d->mIsCompletedTodosCreatedLoaded = loaded;
}

bool ExtendedStorage::isDateLoaded()
{
  return d->mIsDateLoaded;
}

void ExtendedStorage::setIsDateLoaded( bool loaded )
{
  d->mIsDateLoaded = loaded;
}

bool ExtendedStorage::isFutureDateLoaded()
{
  return d->mIsFutureDateLoaded;
}

void ExtendedStorage::setIsFutureDateLoaded( bool loaded )
{
  d->mIsFutureDateLoaded = loaded;
}

bool ExtendedStorage::isJournalsLoaded()
{
  return d->mIsJournalsLoaded;
}

void ExtendedStorage::setIsJournalsLoaded( bool loaded )
{
  d->mIsJournalsLoaded = loaded;
}

bool ExtendedStorage::isCreatedLoaded()
{
  return d->mIsCreatedLoaded;
}

void ExtendedStorage::setIsCreatedLoaded( bool loaded )
{
  d->mIsCreatedLoaded = loaded;
}

bool ExtendedStorage::isGeoDateLoaded()
{
  return d->mIsGeoDateLoaded;
}

void ExtendedStorage::setIsGeoDateLoaded( bool loaded )
{
  d->mIsGeoDateLoaded = loaded;
}

bool ExtendedStorage::isGeoCreatedLoaded()
{
  return d->mIsGeoCreatedLoaded;
}

void ExtendedStorage::setIsGeoCreatedLoaded( bool loaded )
{
  d->mIsGeoCreatedLoaded = loaded;
}

bool ExtendedStorage::isUnreadIncidencesLoaded()
{
  return d->mIsUnreadIncidencesLoaded;
}

void ExtendedStorage::setIsUnreadIncidencesLoaded( bool loaded )
{
  d->mIsUnreadIncidencesLoaded = loaded;
}

bool ExtendedStorage::isInvitationIncidencesLoaded()
{
  return d->mIsInvitationIncidencesLoaded;
}

void ExtendedStorage::setIsInvitationIncidencesLoaded( bool loaded )
{
  d->mIsInvitationIncidencesLoaded = loaded;
}

#if 0
void ExtendedStorage::ExtendedStorageObserver::storageModified( ExtendedStorage *storage,
                                                        const QString &info )
{
  Q_UNUSED( storage );
  Q_UNUSED( info );
}

void ExtendedStorage::ExtendedStorageObserver::storageProgress( ExtendedStorage *storage,
                                                        const QString &info )
{
  Q_UNUSED( storage );
  Q_UNUSED( info );
}

void ExtendedStorage::ExtendedStorageObserver::storageFinished( ExtendedStorage *storage,
                                                        bool error, const QString &info )
{
  Q_UNUSED( storage );
  Q_UNUSED( error );
  Q_UNUSED( info );
}
#endif

void ExtendedStorage::registerObserver( ExtendedStorageObserver *observer )
{
  if ( !d->mObservers.contains( observer ) ) {
    d->mObservers.append( observer );
  }
}

void ExtendedStorage::unregisterObserver( ExtendedStorageObserver *observer )
{
  d->mObservers.removeAll( observer );
}

void ExtendedStorage::setModified( const QString &info, bool mySave )
{
  d->mIsOwnSave = mySave;
  if (!mySave) {
    // Clear all smart loading variables
    d->mStart = QDate();
    d->mEnd = QDate();
    d->mIsUncompletedTodosLoaded = false;
    d->mIsCompletedTodosDateLoaded = false;
    d->mIsCompletedTodosCreatedLoaded = false;
    d->mIsGeoDateLoaded = false;
    d->mIsGeoCreatedLoaded = false;
    d->mIsUnreadIncidencesLoaded = false;
    d->mIsInvitationIncidencesLoaded = false;
  }

  foreach ( ExtendedStorageObserver *observer, d->mObservers ) {
    observer->storageModified( this, info );
  }
}

void ExtendedStorage::setProgress( const QString &info )
{
  foreach ( ExtendedStorageObserver *observer, d->mObservers ) {
    observer->storageProgress( this, info );
  }
}

void ExtendedStorage::setFinished( bool error, const QString &info )
{
  foreach ( ExtendedStorageObserver *observer, d->mObservers ) {
    observer->storageFinished( this, error, info );
  }
}

bool ExtendedStorage::addNotebook( const Notebook::Ptr &nb, bool signal )
{
#if defined(HAVE_UUID_UUID_H)
  uuid_t uuid;
  char suuid[64];
  if ( uuid_parse( nb->uid().toLatin1().data(), uuid ) ) {
    // Cannot accept this id, create better one.
    uuid_generate_random( uuid );
    uuid_unparse( uuid, suuid );
    nb->setUid( QString( suuid ) );
  }
#else
#ifdef __GNUC__
  #warning no uuid support. It is mandatory :)
#endif
#endif

  if ( !nb || d->mNotebooks.contains( nb->uid() ) ) {
    return false;
  }

  if ( !calendar()->addNotebook( nb->uid(), nb->isVisible() ) ) {
    kError() << "cannot add notebook" << nb->uid() << "to calendar";
    return false;
  }

  if ( !modifyNotebook( nb, DBInsert, signal ) ) {
    calendar()->deleteNotebook( nb->uid() );
    return false;
  }
  d->mNotebooks.insert( nb->uid(), nb );

  return true;
}

bool ExtendedStorage::updateNotebook( const Notebook::Ptr &nb )
{
  if ( !nb || !d->mNotebooks.contains( nb->uid() ) ||
       d->mNotebooks.value( nb->uid() ) != nb ) {
    return false;
  }

  if ( !calendar()->updateNotebook( nb->uid(), nb->isVisible() ) ) {
    kError() << "cannot update notebook" << nb->uid() << "in calendar";
    return false;
  }
  if (d->mNotebooks.value(nb->uid())->isVisible() != nb->isVisible())
    return modifyNotebook( nb, DBUpdateVisibility );
  else
    return modifyNotebook( nb, DBUpdate );
}

bool ExtendedStorage::deleteNotebook( const Notebook::Ptr &nb, bool onlyMemory )
{
  if ( !nb || !d->mNotebooks.contains( nb->uid() ) ) {
    return false;
  }

  if ( !modifyNotebook( nb, DBDelete ) ) {
    return false;
  }

  // delete all notebook incidences from calendar
  if ( !onlyMemory ) {
    Incidence::List list;
    Incidence::List::Iterator it;
    if ( allIncidences( &list, nb->uid() ) ) {
      kDebug() << "deleting" << list.size() << "notes of notebook" << nb->name();
      for ( it = list.begin(); it != list.end(); ++it ) {
        load ( (*it)->uid(), (*it)->recurrenceId() );
      }
      for ( it = list.begin(); it != list.end(); ++it ) {
        Incidence::Ptr toDelete = calendar()->incidence( (*it)->uid(), (*it)->recurrenceId() );
        calendar()->deleteIncidence( toDelete );
      }
      if ( !list.isEmpty() ) {
        save();
      }
    } else {
      kError() << "error when loading incidences for notebook" << nb->uid();
      return false;
    }
  }

  if ( !calendar()->deleteNotebook( nb->uid() ) ) {
    kError() << "cannot delete notebook" << nb->uid() << "from calendar";
    return false;
  }

  d->mNotebooks.remove( nb->uid() );

  if ( d->mDefaultNotebook == nb ) {
    d->mDefaultNotebook = Notebook::Ptr();
    // setting new default
    Notebook::Ptr defaultnb = notebook( QString("11111111-2222-3333-4444-555555555555") ); // first trying personal
    if (defaultnb)
      setDefaultNotebook(defaultnb);
    else {
      Notebook::List visibleNotebooks, invisibleNotebooks;
      foreach (const Notebook::Ptr& nb, notebooks()) {
	if ((nb->uid() != "8a2d504d-c53d-4ae8-acfd-78f99706a563") && // ignoring chinese calendars
	    (nb->uid() != "6e50a47f-60c7-4060-bb82-8d2134a013c6")) {
	  if (!nb->isHidden()) {
	    if (nb->isVisible())
	      visibleNotebooks.append(nb);
	    else
	      invisibleNotebooks.append(nb);
	  }
	}
      }

      if ( !visibleNotebooks.isEmpty() )
	setDefaultNotebook(visibleNotebooks.at(0)); // setting first visible as new default notebook
      else if ( !invisibleNotebooks.isEmpty() )
	setDefaultNotebook(invisibleNotebooks.at(0)); // setting first invisible as new default notebook
    }
  }

  return true;
}

bool ExtendedStorage::setDefaultNotebook( const Notebook::Ptr &nb )
{
  if ( !nb || !d->mNotebooks.contains( nb->uid() ) ) {
    return false;
  }

  if (d->mDefaultNotebook) {
    d->mDefaultNotebook->setIsDefault( false );
    if ( !modifyNotebook( d->mDefaultNotebook, DBUpdate, false ) ) {
      return false;
    }
  }

  d->mDefaultNotebook = nb;
  d->mDefaultNotebook->setIsDefault( true );
  if ( !modifyNotebook( d->mDefaultNotebook, DBUpdate ) ) {
    return false;
  }

  Calendar::Ptr cal ( calendar() );

  if ( cal ) {
    calendar()->setDefaultNotebook( nb->uid() );
  }

  return true;
}

Notebook::Ptr ExtendedStorage::defaultNotebook()
{
  if ( d->mDefaultNotebook ) {
    return d->mDefaultNotebook;
  } else {
    return Notebook::Ptr();
  }
}

Notebook::List ExtendedStorage::notebooks()
{
  return d->mNotebooks.values();
}

Notebook::Ptr ExtendedStorage::notebook( const QString &uid )
{
  if ( d->mNotebooks.contains( uid ) ) {
    return d->mNotebooks.value( uid );
  } else {
    return Notebook::Ptr();
  }
}

void ExtendedStorage::setValidateNotebooks( bool validateNotebooks )
{
  d->mValidateNotebooks = validateNotebooks;
}

bool ExtendedStorage::validateNotebooks()
{
  return d->mValidateNotebooks;
}

bool ExtendedStorage::isValidNotebook( const QString &notebookUid )
{
  Notebook::Ptr nb = notebook( notebookUid );
  if ( !nb.isNull() ) {
    if ( nb->isRunTimeOnly() || nb->isReadOnly() ) {
      kDebug() << "notebook" << notebookUid << "isRunTimeOnly or isReadOnly";
      return false;
    }
  } else if ( validateNotebooks() ) {
    kDebug() << "notebook" << notebookUid << "is not valid for this storage";
    return false;
  } else if ( calendar()->hasValidNotebook( notebookUid ) ) {
    kDebug() << "notebook" << notebookUid << "is saved by another storage";
    return false;
  }
  return true;
}

bool ExtendedStorage::isOwnSave()
{
  return d->mIsOwnSave;
}

void ExtendedStorage::resetAlarms( const Incidence::Ptr &incidence )
{
  resetAlarms(Incidence::List(1, incidence));
}

void ExtendedStorage::resetAlarms( const Incidence::List &incidences )
{
#if defined(TIMED_SUPPORT)
  Timed::Interface timed;
  KDateTime now = KDateTime::currentLocalDateTime();
  // list of all timed events
  Timed::Event::List events;

  int alarmNumber(0);

  clearAlarms( incidences );

  foreach(const Incidence::Ptr incidence, incidences) {

    if ( Incidence::StatusCanceled == incidence->status() ) {
        continue;
    }

    // check if the incidence notebook is active
    const QString &notebookUid = calendar()->notebook( incidence->uid() );
    const Notebook::Ptr &nb = notebook(notebookUid);
    if (nb && nb->isVisible()) {
      alarmNumber += d->setAlarms( incidence, events, now );
    }
  }
  //Add all alarms at once
  if ( alarmNumber > 0 ) {
    try
      {
	QDBusReply < QList<QVariant> > reply = timed.add_events_sync( events );
	if ( reply.isValid() ) {
	  foreach (QVariant v, reply.value()) {
            bool ok = true;
            uint cookie = v.toUInt(&ok);
            if (ok && cookie) {
	      kDebug() << "added alarm: " << cookie;
            } else {
	      kError() << "failed to add alarm";
            }
	  }
	} else {
	  kError() << "failed to add alarms: " << reply.error().message();
	}
      }
    catch(Timed::Exception &e)
      {
	kError() << "timed exception caught, failed to add alarm: " << e.what();
      }
  }

#else
  foreach(const Incidence::Ptr incidence, incidences) {
    clearAlarms( incidence );
    setAlarms( incidence);
  }
#endif
}

void ExtendedStorage::setAlarms( const Incidence::Ptr &incidence )
{
#if defined(TIMED_SUPPORT)
  // check if the incidence notebook is active
  const QString &notebookUid = calendar()->notebook( incidence->uid() );
  const Notebook::Ptr &nb = notebook(notebookUid);
  if (nb && !nb->isVisible()) {
    kDebug() << "setAlarms: inactive notebook: " << notebookUid;
    return;
  }

  Timed::Interface timed;
  int alarmNumer = 0;

  if ( !timed.isValid() ) {
    kError() << "cannot set alarm for incidence: "
             << "alarm interface is not valid" << timed.lastError();
    return;
  }

  // list of all timed events
  Timed::Event::List events;

  Alarm::List alarms = incidence->alarms();
  foreach ( const Alarm::Ptr alarm, alarms ) {
    if ( !alarm->enabled() ) {
      continue;
    }

    if ( alarm->type() != Alarm::Display && alarm->type() != Alarm::Audio ) {
      continue;
    }

    KDateTime now = KDateTime::currentLocalDateTime();
    KDateTime alarmTime = alarm->nextTime( now, true );
    if ( !alarmTime.isValid() ) {
      continue;
    }

    if ( now.addSecs( 60 ) > alarmTime ) {
      // don't allow alarms at the same minute -> take next alarm if so
      alarmTime = alarm->nextTime( now.addSecs( 60 ), true );
      if ( !alarmTime.isValid() ) {
        continue;
      }
    }
    Timed::Event& e = events.append();
    alarmNumer++;
    e.setUserModeFlag();
    if ( alarmTime.isUtc() ) {
      e.setTicker( alarmTime.toTime_t() );
    } else {
      e.setTicker( alarmTime.toUtc().toTime_t() );
    }
    // The code'll crash (=exception) iff the content is empty. So
    // we have to check here.
    QString s;

    s = incidence->summary();
    // Timed braindeath: Required field, BUT if empty, it asserts
    if ( s.isEmpty() ) {
      s = ' ';
    }
    e.setAttribute( "TITLE", s );
    e.setAttribute( "PLUGIN", "libCalendarReminder" );
    e.setAttribute( "APPLICATION", "libextendedkcal" );
    //e.setAttribute( "translation", "organiser" );
    // This really has to exist or code is badly broken
    Q_ASSERT( !incidence->uid().isEmpty() );
    e.setAttribute( "uid", incidence->uid() );
#ifndef QT_NO_DEBUG_OUTPUT //Helps debuggin
    e.setAttribute( "alarmtime", alarmTime.toString() );
#endif
    if ( !incidence->location().isEmpty() ) {
      e.setAttribute( "location", incidence->location() );
    }
    if ( incidence->recurs() ) {
      e.setAttribute( "recurs", "true" );
    }

    // TODO - consider this how it should behave for recurrence
    if ( ( incidence->type() == Incidence::TypeTodo ) ) {
      Todo::Ptr todo = incidence.staticCast<Todo>();

      if ( todo->hasDueDate() ) {
        e.setAttribute( "time", todo->dtDue( true ).toString() );
      }
      e.setAttribute( "type", "todo" );
    } else if ( incidence->dtStart().isValid() ) {
      e.setAttribute( "time", incidence->dtStart().toString() );
      e.setAttribute( "type", "event" );
    }

    if ( incidence->hasRecurrenceId() ) {
      e.setAttribute( "recurrenceId", incidence->recurrenceId().toString() );
    }
    e.setAttribute( "notebook", notebookUid );

    Timed::Event::Button &s15 = e.addButton();
    s15.setAttribute( "snooze_value", "15" );
    s15.setSnooze( 900 );
    s15.setAttribute( "label", "Snooze 15 minutes" );

    Timed::Event::Button &s10 = e.addButton();
    s10.setAttribute( "snooze_value", "10" );
    s10.setSnooze( 600 );
    s10.setAttribute( "label", "Snooze 10 minutes" );

    Timed::Event::Button &s05 = e.addButton();
    s05.setAttribute( "snooze_value", "5" );
    s05.setSnooze( 300 );
    s05.setAttribute( "label", "Snooze 5 minutes" );

    Timed::Event::Button &open = e.addButton();
    open.setAttribute( "label", "Close" );

    e.hideSnoozeButton1();
    e.setAlignedSnoozeFlag();
    e.setTriggerIfMissedFlag();
  }

  if (alarmNumer > 0) {
    try
      {
	QDBusReply < QList<QVariant> > reply = timed.add_events_sync( events );
	if ( reply.isValid() ) {
	  foreach (QVariant v, reply.value()) {
	    bool ok = true;
	    uint cookie = v.toUInt(&ok);
	    if (ok && cookie) {
	      kDebug() << "added alarm: " << cookie;
	    } else {
	      kError() << "failed to add alarm";
	    }
	  }
	} else {
	  kError() << "failed to add alarms: " << reply.error().message();
	}
      }
    catch(Timed::Exception &e)
      {
	kError() << "timed exception caught, failed to add alarms: " << e.what();
      }
  } else {
      kDebug() << "No alarms to send";
  }
#else
  Q_UNUSED( incidence );
#endif
}

void ExtendedStorage::clearAlarms( const Incidence::Ptr &incidence )
{
    clearAlarms(Incidence::List(1, incidence));
}

void ExtendedStorage::clearAlarms( const KCalCore::Incidence::List &incidences )
{
#if defined(TIMED_SUPPORT)
  QList<uint32_t> cookies;

  Timed::Interface timed;
  if ( !timed.isValid() ) {
    kError() << "cannot clear alarms for the list. alarm interface is not valid" << timed.lastError();
    return;
  }

  QMap<QString,QString> map;
  map["APPLICATION"] = "libextendedkcal";
  QDBusReply<QList<uint> > qreply;
  try
    {
      qreply = timed.get_cookies_by_attributes_sync(map);

    }
  catch(Timed::Exception &e)
    {
      kError() << "timed exception caught, cannot get cookies of application" << e.what();
      return;
    }

  const QList<uint> &allCookies = qreply.value();
  if (allCookies.isEmpty())
    return;

  QDBusReply<QMap<uint,QMap<QString,QString> > > areply;
  try
    {
      areply = timed.get_attributes_by_cookies_sync(allCookies) ;
    }
  catch(Timed::Exception &e)
    {
      kError() << "timed exception caught, cannot get attributes by cookies" << e.what();
      return;
    }

  const QMap<uint,QMap<QString,QString> > &attr = areply.value();

  foreach(const Incidence::Ptr incidence, incidences) {
    const QString &uid = incidence->uid();

    // Check if we have the alarm in timed corresponding to the current incidence
    uint cookie = 0;
    QMap<uint,QMap<QString,QString> >::const_iterator end = attr.end();
    QMap<uint,QMap<QString,QString> >::const_iterator it = attr.begin();
    for (; it != end; ++it) {
        const QMap<QString,QString> &value = it.value();
        if (value["uid"] == uid) {
            const QString &timedRecurrenceId = value["recurrenceId"];
            if (incidence->hasRecurrenceId()) {
                const QString &recurrenceId = incidence->recurrenceId().toString();
                if (timedRecurrenceId == recurrenceId) {
                    cookie = it.key();
                    break;
                }
            }
            else {
                if (timedRecurrenceId.isEmpty()) {
                    cookie = it.key();
                    break;
                }
            }
        }
    }
    if (cookie != 0) {
      cookies.append( cookie );
    }
  }

  if (cookies.isEmpty())
    return;

  try
    {
      QDBusReply<bool> reply = timed.cancel_events_sync( cookies );
      if ( !reply.isValid() || !reply.value() ) {
        kError() << "cannot remove the set of alarms" << reply.value() << timed.lastError() << cookies;
      }
    }
  catch(Timed::Exception &e)
    {
      kError() << "timed exception caught, cannot remove the set of alams: " << e.what();
    }

#else
  Q_UNUSED( incidences );
#endif
}

void ExtendedStorage::clearAlarms( const QString &nb )
{
#if defined(TIMED_SUPPORT)
  Incidence::List list;
  if ( allIncidences( &list, nb ) ) {
    clearAlarms(list);
  } else {
    kError() << "error when loading incidences for notebook" << nb;
  }
#else
  Q_UNUSED( nb );
#endif
}

Incidence::Ptr ExtendedStorage::checkAlarm( const QString &uid, const QString &recurrenceId,
                                            bool loadAlways )
{
  KDateTime rid;

  if ( !recurrenceId.isEmpty() ) {
    rid = KDateTime::fromString( recurrenceId );
  }
  Incidence::Ptr incidence = calendar()->incidence( uid, rid );
  if ( !incidence || loadAlways ) {
    load( uid, rid );
    incidence = calendar()->incidence( uid, rid );
  }
  if ( incidence && incidence->hasEnabledAlarms() ) {
    // Return incidence if it exists and has active alarms.
    return incidence;
  }
  return Incidence::Ptr();
}


Notebook::Ptr ExtendedStorage::createDefaultNotebook( QString name, QString color )
{
  QString uid;
  if (name.isEmpty())
    name = "Default";
  if (color.isEmpty())
    color = "#0000FF";

  Notebook::Ptr nbDefault = Notebook::Ptr( new Notebook(name, QString() ) );
  nbDefault->setColor( color );
  addNotebook(nbDefault, false);
  setDefaultNotebook(nbDefault);
  return nbDefault;
}


#if defined(TIMED_SUPPORT)
int ExtendedStorage::Private::setAlarms( const Incidence::Ptr &incidence, Timed::Event::List &events, const KDateTime &now) {

  Alarm::List alarms = incidence->alarms();

  int alarmNumber(0);

  foreach ( const Alarm::Ptr alarm, alarms ) {
    if ( !alarm->enabled() ) {
      continue;
    }

    if ( alarm->type() != Alarm::Display && alarm->type() != Alarm::Audio ) {
      continue;
    }

    KDateTime alarmTime = alarm->nextTime( now, true );
    if ( !alarmTime.isValid() ) {
      continue;
    }

    if ( now.addSecs( 60 ) > alarmTime ) {
      // don't allow alarms at the same minute -> take next alarm if so
      alarmTime = alarm->nextTime( now.addSecs( 60 ), true );
      if ( !alarmTime.isValid() ) {
        continue;
      }
    }
    Timed::Event& e = events.append();
    e.setUserModeFlag();
    e.setBackupFlag();

    if ( alarmTime.isUtc() ) {
      e.setTicker( alarmTime.toTime_t() );
    } else {
      e.setTicker( alarmTime.toUtc().toTime_t() );
    }
    // The code'll crash (=exception) iff the content is empty. So
    // we have to check here.
    QString s;

    s = incidence->summary();
    // Timed braindeath: Required field, BUT if empty, it asserts
    if ( s.isEmpty() ) {
      s = ' ';
    }
    e.setAttribute( "TITLE", s );
    e.setAttribute( "PLUGIN", "libCalendarReminder" );
    e.setAttribute( "APPLICATION", "libextendedkcal" );
    //e.setAttribute( "translation", "organiser" );
    // This really has to exist or code is badly broken
    Q_ASSERT( !incidence->uid().isEmpty() );
    e.setAttribute( "uid", incidence->uid() );
#ifndef QT_NO_DEBUG_OUTPUT //Helps debuggin
    e.setAttribute( "alarmtime", alarmTime.toString() );
#endif
    if ( !incidence->location().isEmpty() ) {
      e.setAttribute( "location", incidence->location() );
    }
    if ( incidence->recurs() ) {
      e.setAttribute( "recurs", "true" );
    }

    // TODO - consider this how it should behave for recurrence
    if ( ( incidence->type() == Incidence::TypeTodo ) ) {
      Todo::Ptr todo = incidence.staticCast<Todo>();

      if ( todo->hasDueDate() ) {
        e.setAttribute( "time", todo->dtDue( true ).toString() );
      }
      e.setAttribute( "type", "todo" );
    } else if ( incidence->dtStart().isValid() ) {
      e.setAttribute( "time", incidence->dtStart().toString() );
      e.setAttribute( "type", "event" );
      if (incidence->allDay()) {
        e.setAttribute( "allday", "true" );
      }
    }

    if ( incidence->hasRecurrenceId() ) {
      e.setAttribute( "recurrenceId", incidence->recurrenceId().toString() );
    }
    e.setAttribute( "notebook", mCalendar->notebook( incidence->uid() ) );

    Timed::Event::Button &s15 = e.addButton();
    s15.setAttribute( "snooze_value", "15" );
    s15.setSnooze( 900 );
    s15.setAttribute( "label", "Snooze 15 minutes" );

    Timed::Event::Button &s10 = e.addButton();
    s10.setAttribute( "snooze_value", "10" );
    s10.setSnooze( 600 );
    s10.setAttribute( "label", "Snooze 10 minutes" );

    Timed::Event::Button &s05 = e.addButton();
    s05.setAttribute( "snooze_value", "5" );
    s05.setSnooze( 300 );
    s05.setAttribute( "label", "Snooze 5 minutes" );

    Timed::Event::Button &open = e.addButton();
    open.setAttribute( "label", "Close" );

    e.hideSnoozeButton1();
    e.setAlignedSnoozeFlag();
    e.setTriggerIfMissedFlag();
    ++alarmNumber;
  }
  return alarmNumber;
}
#endif

void ExtendedStorage::setNotebookVisibility(const QString& id, bool visible)
{
    Notebook::Ptr nb = notebook(id);
    Q_ASSERT(nb);

    bool visibilityChanged = visible != nb->isVisible();

    nb->setIsVisible(visible);
    updateNotebook(nb);

    if (visibilityChanged == true) {
        // We can set notebook as default only if it is visible
        // otherwize we need to reset default notebook
        // in some case it might be set to the same notebook
        // if there are no other active notebooks and this
        // notebook was created first.
        if (nb->isDefault() && visible == false) {
            resetDefaultNotebook();
        }

        // Set alarms if notebook is visible
        // and clear if notebook is invisible
        if (visible == true) {
            KCalCore::Incidence::List list;
            allIncidences(&list,id);
            resetAlarms(list);
        } else {
            clearAlarms(id);
        }
    }
}

void ExtendedStorage::resetDefaultNotebook()
{
    Notebook::Ptr currentDefaultNb = defaultNotebook();
    Notebook::List visibleNotebooks, invisibleNotebooks;

    // Next notebook selection algorithm works by first dividing
    // notebooks to two groups: visible and invisible. Each group is
    // then sorted. We then try to select the first notebook from the
    // visible group, but will not accept the notebook that is the
    // current default. Next we process the invisible notebooks
    // group. Finally, if there is just the current default notebook,
    // that one is ratained as the default.

    foreach (const Notebook::Ptr& nb, notebooks()) {
        if (!nb->isHidden() && !nb->isReadOnly()) {
            if (nb->isVisible()) {
                visibleNotebooks.append(nb);
            } else {
                invisibleNotebooks.append(nb);
            }
        }
    }

    qStableSort(visibleNotebooks.begin(), visibleNotebooks.end(),
                notebookModifiedEarlierThan);

    foreach (const Notebook::Ptr& nb, visibleNotebooks) {
        if (currentDefaultNb == 0 || currentDefaultNb->uid() != nb->uid()) {
            setDefaultNotebook(nb);
            return;
        }
    }

    qStableSort(invisibleNotebooks.begin(), invisibleNotebooks.end(),
                notebookModifiedEarlierThan);

    // Set first invisible
    if (!invisibleNotebooks.isEmpty()) {
        setDefaultNotebook(*(invisibleNotebooks.begin()));
        return;
    }

    // there is only the current default notebook
    Q_ASSERT(currentDefaultNb);
    setDefaultNotebook(currentDefaultNb);
}

bool ExtendedStorage::notebookModifiedEarlierThan(const Notebook::Ptr& nb1,
                                                  const Notebook::Ptr& nb2)
{
    return nb1->creationDate() < nb2->creationDate();
}
